/*
 * com/ngon/io/TextFileRunner.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.io;

import com.ngon.app.App;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.util.Vector;

/**
 * class TextFileRunner
 */
final public class TextFileRunner extends Thread {
    
    private String req;
    private String lastReq;
    private App app;
    private String[] lines;
    private Thread thisThread;
    private boolean isVerbose;
    
    public TextFileRunner(App app) {
        this.app = app;
        setDaemon(true);
    }
    
    public synchronized void setRequest(String req) {
        this.req = req;
        this.lastReq = req;
        this.lines = null;
        notify();
        if (isVerbose) {
            System.err.println("TextFileRunner.setRequest() : " + req);
        }
    }
    
    public synchronized String getRequest() {
        return this.req;
    }
    
    public void run() {
        thisThread = Thread.currentThread();
        while (this == thisThread) {
            try {
                if (req == null) {
                    synchronized(this) {
                        while (req == null) {
                            wait();
                        }
                    }
                }
            } catch (InterruptedException e) {}
            
            if (this != thisThread) {
                return;
            }
            
            InputStream pipe = app.getDataStream(req);
            if (pipe == null) {
                if (isVerbose) {
                    System.err.println(
                        "TextFileRunner.run() : " +
                        "Couldn't retrieve data stream from " + req);
                }
                // Reset.
                synchronized(this) {
                    req = null;
                }
                continue;
            }
            
            try {
                BufferedReader reader = 
                    new BufferedReader(new InputStreamReader(pipe));
                Vector v = new Vector();
                String s;
                while ((s = reader.readLine()) != null) {
                    v.addElement(s);
                }
                reader.close();
                pipe.close();
                int len = v.size();
                if (len > 0) {
                    synchronized(this) {
                        lines = new String[len];
                        v.copyInto(lines);
                        if (isVerbose) {
                            System.err.println(
                                "TextFileRunner.run() : copied " + len + 
                                " lines");
                        }
                    }
                    v.removeAllElements();
                }
                // Reset src.
                synchronized(this) {
                    req = null;
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
    
    public synchronized String[] getLines() {
        return lines;
    }
    
    public synchronized void clear() {
        this.lines = null;
    }
    
    public synchronized void resendRequest() {
        this.req = new String(lastReq);
        if (isVerbose) {
            System.err.println(
                "TextFileRunner.resendRequest() : " + req);
        }
        notify();
    }
    
    public void setVerbose(boolean isVerbose) {
        this.isVerbose = isVerbose;
    }
    
    public void kill() {
        if (thisThread != null) {
            Thread moribund = thisThread;
            thisThread = null;
            moribund.interrupt();
        }
    }
}
