/*
 * com/ngon/tap/TapApp.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.app.App;
import com.ngon.gfx.PixelSurface;
import com.ngon.gfx.RGB;
import com.ngon.mpeg.MPEGFrame;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.sql.Connection;

/**
 * class TapApp
 */
public abstract class TapApp extends App implements KeyListener {
    
    final static Dimension FRAME    = new Dimension(352,240);
    final static Dimension Z1       = new Dimension(424,312);
    final static Dimension Z2       = new Dimension(840,616);
    
    final static int SLEEP          = 100;
    
    final static int[] CLUSTER_RGB  = {
        0xCC3333, 0x33CC33, 0x3333CC, 0xCCCC00, 0x00CCCC, 0xCC00CC, 0x999999
    };
    
    protected int zoom = 2;
    protected Dimension appDim;
    protected Point origin;
    
    protected Connection dbc;
    protected boolean waiting;
    
    static public void main(String title, TapApp a, String[] args) {
        a.getApplicationParameters(args);
        a.initCoords();
        
        Frame f = new Frame(title);
        f.addWindowListener(
            new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    System.exit(0);
                }
            });
        f.setLayout(null);
        f.setBackground(Color.black);
        f.setBounds(0, 0, a.appDim.width+24, a.appDim.height+36);
        f.show();
        
        Insets insets = f.getInsets();
        a.setLocation(insets.left, insets.top);
        a.setSize(a.appDim.width, a.appDim.height);
        a.initCommon();
        
        f.add(a);
        f.addKeyListener(a);
        a.addKeyListener(a);
        a.start();
    }
    
    public void init() {
        super.init();
        getAppletParameters();
        initCoords();
        initCommon();
        addKeyListener(this);
    }
    
    protected void initCoords() {
        if (zoom == 1) {
            appDim = new Dimension(Z1);
        } else if (zoom == 2) {
            appDim = new Dimension(Z2);
        } else {
            System.err.println("Zoom value " + zoom + " not supported.");
            kill();
        }
        origin = new Point((appDim.width  - zoom * FRAME.width)  / 2,
                           (appDim.height - zoom * FRAME.height) / 2);
    }
    
    protected abstract void getApplicationParameters(String[] args);
    protected abstract void getAppletParameters();
    protected abstract void initCommon();
    
    /* ========================================================================
        PAINT OPS
       ========================================================================
     */
    protected void paintFrame(Clip clip, PixelSurface pixels) {
        MPEGFrame frame = clip.getFrame();
        if (frame == null) {
            System.err.println("App.paintFrame() : MPEGFrame is null");
            return;
        }
        int[] rgbField = frame.getRGB();
        Dimension d = clip.getFrameDimensions();
        pixels.paintField(origin.x, origin.y, rgbField, d.width, d.height);
    }
    
    protected void paintMV(int x, int y, Point mv, int alpha, 
                           PixelSurface pixels) {
        if (mv == null) {
            return;
        }
        paintMV(x, y, mv.x, mv.y, alpha, pixels);
    }
    
    protected void paintMV(int x, int y, int mvX, int mvY, int alpha,
                           PixelSurface pixels) {
        int dx = mvX * zoom;
        int dy = mvY * zoom;
        pixels.drawWuLine(x - 2, y, x + 3, y, 0x9999FF, alpha);
        pixels.drawWuLine(x, y - 2, x, y + 3, 0x9999FF, alpha);
        pixels.drawWuLine(x, y, x + dx, y + dy, RGB.WHT, alpha);
    }
    
    protected void paintAvg(int x, int y, int dx, int dy, int rgb, int crossRGB,
                            int mvRGB, PixelSurface pixels) {
        pixels.drawWuLine(x - 5, y - 5, x + 6, y - 5, rgb);
        pixels.drawWuLine(x - 5, y - 5, x - 5, y + 6, rgb);
        pixels.drawWuLine(x + 5, y - 5, x + 5, y + 6, rgb);
        pixels.drawWuLine(x - 5, y + 5, x + 6, y + 5, rgb);
        pixels.drawWuLine(x - 4, y - 4, x + 5, y - 4, rgb);
        pixels.drawWuLine(x - 4, y - 4, x - 4, y + 5, rgb);
        pixels.drawWuLine(x + 4, y - 4, x + 4, y + 5, rgb);
        pixels.drawWuLine(x - 4, y + 4, x + 5, y + 4, rgb);
        pixels.drawWuLine(x - 6, y, x + 7, y, crossRGB);
        pixels.drawWuLine(x - 5, y - 1, x + 6, y - 1, crossRGB);
        pixels.drawWuLine(x - 5, y + 1, x + 6, y + 1, crossRGB);
        pixels.drawWuLine(x, y - 6, x, y + 7, crossRGB);
        pixels.drawWuLine(x - 1, y - 5, x - 1, y + 6, crossRGB);
        pixels.drawWuLine(x + 1, y - 5, x + 1, y + 6, crossRGB);
        pixels.drawWuLine(x, y, x + dx, y + dy, mvRGB);
        pixels.drawWuLine(x, y, x + dx, y + dy, mvRGB);
        pixels.drawWuLine(x, y, x + dx, y + dy, mvRGB);
    }
    
    protected void paintBlockOutline(int x, int y, int rgb, 
                                     PixelSurface pixels) {
        int l = 16 * zoom - 1;
        pixels.drawWuLine(x+1, y, x+l, y, rgb);
        pixels.drawWuLine(x, y+1, x, y+l, rgb);
        pixels.drawWuLine(x+l, y+1, x+l, y+l, rgb);
        pixels.drawWuLine(x+1, y+l, x+l, y+l, rgb);
        pixels.drawWuLine(x+2, y+1, x+l-1, y+1, rgb);
        pixels.drawWuLine(x+1, y+2, x+1, y+l-1, rgb);
        pixels.drawWuLine(x+l-1, y+2, x+l-1, y+l-1, rgb);
        pixels.drawWuLine(x+2, y+l-1, x+l-1, y+l-1, rgb);
    }
    
    /* ========================================================================
        KEY OPS
       ========================================================================
     */
    public void keyTyped(KeyEvent e) {}
    
    public void keyPressed(KeyEvent e) {
        if (isRunningAsApplet) {
            return;
        }
        
        synchronized(this) {
            switch (e.getKeyCode()) {
            case KeyEvent.VK_ESCAPE: 
            case KeyEvent.VK_Q: 
                kill();
                return;
            case KeyEvent.VK_SPACE:
                waiting = false;
                break;
            }
        }
    }

    public void keyReleased(KeyEvent e) {}
}
