/*
 * com/ngon/tap/TapEvol.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.app.App;
import com.ngon.gfx.ImageBuffer;
import com.ngon.gfx.RGB;
import com.ngon.gfx.PixelDrawer;
import com.ngon.gfx.PixelSurface;
import com.ngon.gfx.ZoomPixelSurface;
import com.ngon.mpeg.MPEGFrame;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * class TapEvol
 */
final public class TapEvol extends TapApp implements MouseListener,
                                                     MouseMotionListener {
    
    final static int WAIT_SLEEP         = 100;
    final static int MIN_SLEEP          = 5;
    final static int DT                 = 50;
    
    final static int TOO_MANY_EVOL_FAILURES = 60000 / DT; // Wait 60 sec.
    
    // Frame insets.
    final static int L                  = 0;
    final static int T                  = 0;
    final static int R                  = 0;
    final static int B                  = 0;
    
    final static int PANEL_DIM_COUNT    = 7000 / DT;
    
    final static int PANEL_HIDE         = 0;
    final static int PANEL_SHOW         = 1;
    final static int PANEL_HIDING       = 2;
    final static int PANEL_SHOWING      = 3;
    
    final static int PANEL_W            = 186;
    final static int PANEL_INV_SLOPE    = 3;
    final static int PANEL_MAX_VEL      = 10;
    final static int PANEL_ACCEL        = 2;
    final static int PANEL_ACCEL_DIST   = PANEL_MAX_VEL * PANEL_MAX_VEL *
                                          PANEL_INV_SLOPE /
                                          (2 * PANEL_ACCEL);
    
    final static int PANEL_X1           = 210; // Leftmost dist from rhs.
    final static int PANEL_X2           = 6;   // Rightmost dist from rhs.
    final static int PANEL_ACTIVATE_Y   = 81;
    final static int PANEL_ACTIVATE_W   = 32;
    final static int PANEL_ACTIVATE_H   = 93;
    final static int PANEL_STAY_Y       = 0;
    final static int PANEL_STAY_W       = 280;
    
    final static int PANEL_ALPHA_INCR   = 25;
    final static int PANEL_ALPHA_DECR   = 10;
    final static int MIN_PANEL_ALPHA    = 55;
    final static int MAX_PANEL_ALPHA    = 255;
    
    // Ctrls
    final static int BEAT_PERIOD        = 0;
    final static int FADE_PERIOD        = 1;
    final static int TAP0_ALPHA         = 2;
    final static int ALPHA_RATIO        = 3;
    final static int ERASER             = 4;
    final static int N_TAPS             = 5;
    final static int MV_ALPHA           = 6;
    final static int AGG_ALPHA          = 7;
    
    // Connection indicator
    final static int CONN_GOOD          = 0;
    final static int CONN_WAIT          = 1;
    final static int CONN_BAD           = 2;
    
    final static int[][] CTRL_SETTINGS = { 
        // Beat periods
        { 250, 300, 350, 400, 450, 500, 600, 700, 800, 900, 1000, 
        1100, 1200, 1300, 1400, 1500, 1600, 1700, 1800, 1900, 2000 },
        // Fade periods
        { 300, 400, 500, 600, 800, 1000, 1200, 1500, 1800, 2200, 2600, 
        3000, 3500, 4000, 5000, 6000, 7000, 8000, 10000, 20000, 100000 },
        // Initial alpha of top tap
        { 255, 238, 221, 204, 187, 170, 153, 
        136, 119, 102, 85, 68, 51, 34, 17, 0 },
        // Alpha ratio between taps
        { 256, 224, 192, 160, 128, 112, 96, 
        80, 72, 64, 48, 40, 32, 24, 16, 8 },
        // Eraser
        { 255, 128, 96, 64, 48, 32, 24, 
        16, 12, 8, 6, 4, 3, 2, 1, 0 },
        // Nr of taps
        { 2, 1 },
        // MV alpha
        { 255, 238, 221, 204, 187, 170, 153, 
        136, 119, 102, 85, 68, 51, 34, 17, 0 },
        // Aggregate alpha
        { 255, 238, 221, 204, 187, 170, 153, 
        136, 119, 102, 85, 68, 51, 34, 17, 0 },
    };
    final static int N_CTRLS            = CTRL_SETTINGS.length;
    final static String[] ICON_IMGS     = {
        "/images/icon-beat-period.gif",
        "/images/icon-fade-period.gif",
        "/images/icon-tap0-alpha.gif",
        "/images/icon-alpha-ratio.gif",
        "/images/icon-eraser.gif",
        "/images/icon-n-taps.gif",
        "/images/icon-mv-alpha.gif",
        "/images/icon-agg-alpha.gif"
    };
    
    final static int CTRL_W             = 21;
    final static int CTRL_H             = 60;
    final static int CTRL_X0            = 11;
    final static int CTRL_Y0            = 28;
    final static int CTRL_DX            = 21;
    final static int CTRL_DY            = 7;
    
    final static int ICON_W             = 14;
    final static int ICON_H             = 14;
    final static int ICON_X0            = 15;
    final static int ICON_Y0            = 10;
    
    final static int TIC_W              = 7;
    final static int TIC_X0             = 7;
    
    final static int TIC_HI_W           = 11;
    final static int TIC_HI_X0          = 5;
    final static int TIC_HI_Y0          = -2;
    
    final static int TIC_GLO_W          = 21;
    final static int TIC_GLO_X0         = 0;
    final static int TIC_GLO_Y0         = -7;
    
    final static int AGG_CIRCLE_W       = 53;
    final static int AGG_CIRCLE_X       = -26;
    final static int AGG_CIRCLE_Y       = -26;
    
    final static int AGG_PTS_IN_W       = 37;
    final static int AGG_PTS_IN_X       = -18;
    final static int AGG_PTS_IN_Y       = -18;
    
    final static int AGG_PTS_OUT_W      = 75;
    final static int AGG_PTS_OUT_X      = -37;
    final static int AGG_PTS_OUT_Y      = -37;
    
    final static int AGG_DISK_W         = 41;
    final static int AGG_DISK_X         = -20;
    final static int AGG_DISK_Y         = -20;
    
    final static int CONN_IND_W         = 8;
    final static int CONN_IND_H         = 8;
    
    final static int ERR_MSG_W          = 260;
    
    final static int PANEL_RGB          = 0x2A415C;
    final static int TIC_RGB            = 0x656E66;
    final static int TIC_HI_RGB         = 0xB8CE97;
    final static int TIC_GLO_RGB        = 0x84FF00;
    final static int AVG_MV_RGB         = 0xFF8151;
    final static int AGG_PTS_RGB        = 0x2F5AF5;
    final static int AGG_DISK_TOP_RGB   = 0x006600;
    final static int AGG_DISK_BOT_RGB   = 0x99FF99;
    final static int ERR_MSG_RGB        = 0xFF0000;
    
    final static int[] CONN_RGB = { 0x00CC00, 0xFFFF00, 0xFF0000 };
    
    // Set some defaults for app params.
    protected String cacheUrl   = "";
    protected String dbHost     = "localhost";
    protected String dbUser     = "victor";
    protected String dbPass     = "";
    private String clipsPath    = "clips";
    private boolean doWait      = false;
    int[] ctrlInit = { 5, 15, 0, 0, 4, 0, 15, 15 };
    
    private Taps taps;
    private int discardTapCount;
    
    private int panelXLeftmost;
    private int panelXRightmost;
    private Rectangle panelActivate;
    private Rectangle panelStay;
    private int panelXIcept1;
    private int panelXIcept2;
    private int panelStayXIcept1;
    private int panelStayXIcept2;
    private int panelState = PANEL_HIDE;
    private int panelAlpha = MIN_PANEL_ALPHA;
    private int panelX;
    private int panelY;
    private int panelV = 0;
    
    private Rectangle erase;
    
    private int[] panelMask;
    private int[] ticMask;
    private int[] ticHiMask;
    private int[] ticGloMask;
    private int[] ticSliderMask;
    private int[] aggCircleMask;
    private int[] aggPtsInFlatMask;
    private int[] aggPtsInTiltLMask;
    private int[] aggPtsInTiltRMask;
    private int[] aggPtsOutFlatMask;
    private int[] aggPtsOutTiltLMask;
    private int[] aggPtsOutTiltRMask;
    private int[] aggDiskTopFlatMask;
    private int[] aggDiskTopTiltLMask;
    private int[] aggDiskTopTiltRMask;
    private int[] aggDiskBotFlatMask;
    private int[] aggDiskBotTiltLMask;
    private int[] aggDiskBotTiltRMask;
    private int[][] iconMasks;
    private int[] connIndMask;
    
    private int[] errMsgMask;
    private int errMsgW;
    
    private Ctrl[] ctrls;
    
    private PixelSurface pixels;
    private PixelSurface composite;
    private Evolver evol;
    private Hashtable clipInfo;
    private ClipSegFactory csFactory;
    
    private boolean isOverPanel;
    private boolean isOverPanelStay;
    private boolean isOverPanelActivate;
    private int mouseInactiveCount;
    
    private int connStatus = CONN_WAIT;
    
    private Dimension frameDim;
    private Dimension blockDim;
    private int blockSz;
    
    static public void main(String[] args) {
        if (args.length < 1) {
            printUsage();
            System.exit(0);
        }
        TapApp.main("Tap Evol", new TapEvol(), args);
    }
    
    static public void printUsage() {
        System.out.println(
            "Usage: java -jar tap-evol.jar <clipsPath> [options]");
        System.out.println("Available options are");
        System.out.println("  -dbHost     Database host");
        System.out.println("  -dbUser     Database user");
        System.out.println("  -zoom       Zoom factor");
        System.out.println("  -wait       Wait between frames");
        System.out.println("  -ctrl<N>    Initial setting of ctrl <N>");
    }
    
    public void start() {
        super.start();
        addMouseListener(this);
        addMouseMotionListener(this);
        requestFocus();
    }

    public void main(int width, int height) {
        // Initial paint.
        composite.copy(pixels);
        paintConnInd(composite);
        update(composite.getPixels());
        
        if (clipInfo == null) {
            // Problem with evol (servlet or database), abort (after paint).
            connStatus = CONN_BAD;
            if (Debug.DEBUG_APPLET) {
                System.err.println(
                    "TapEvol.main() : clipInfo is null ... Aborting.");
            }
        }
        
        long t0, t;
        long tBeat = 0;
        int nEvolFailures = 0;
        
        while (true) {
            if (connStatus == CONN_BAD) {
                composite.copy(pixels);
                paintErrorMsg(composite);
                paintConnInd(composite);
                update(composite.getPixels());
                try {
                    Thread.sleep(2000);
                } catch (InterruptedException e) {}
                continue;
            }
            
            t = t0 = System.currentTimeMillis();
            
            if ((t - tBeat > ctrls[BEAT_PERIOD].getValue()) &&
                 (panelState == PANEL_HIDE || panelState == PANEL_SHOW)) {
                if (evol.evol(taps, ctrls[N_TAPS].getValue(), csFactory)) {
                    tBeat = t;
                    nEvolFailures = 0;
                    connStatus = CONN_GOOD;
                } else {
                    if (nEvolFailures++ > TOO_MANY_EVOL_FAILURES) {
                        System.err.println(
                            "TapEvol.main : TOO MANY EVOL FAILURES");
                        connStatus = CONN_BAD;
                        continue;
                    } else {
                        connStatus = CONN_WAIT;
                    }
                }
            }
            
            // Paint.
            pixels.mixBlack(erase.x, erase.y, erase.width, erase.height,
                            ctrls[ERASER].getValue());
            int count = taps.discardCount;
            int fadePeriod = ctrls[FADE_PERIOD].getValue();
            int alphaRatio = ctrls[ALPHA_RATIO].getValue();
            int fadeFactor = (t - tBeat < fadePeriod) ?
                256 - (((int) (t - tBeat)) << 8) / fadePeriod : 
                0;
            if (fadeFactor < 0) {
                fadeFactor = 0;
            }
            if (fadeFactor > 0) {
                int[] tapFade = new int[taps.size()];
                for (int i = taps.size()-1; i > -1; i--) {
                    tapFade[i] = fadeFactor;
                    if (alphaRatio < 255) {
                        fadeFactor = (fadeFactor * alphaRatio) >> 8;
                    }
                }
                for (int i = 0; i < taps.size(); i++) {
                    if (tapFade[i] == 0) {
                        continue;
                    }
                    Tap tap = (Tap) taps.elementAt(i);
                    paintBlocks(tap.clus, tapFade[i], pixels);
                    int rgb = CLUSTER_RGB[count++ % CLUSTER_RGB.length];
                    paintAgg(tap.clus, rgb, tapFade[i], pixels);
                }
            }
            
            // Determine panel state.  Copy mouse params in case they change.
            boolean isOverPanelActivate = this.isOverPanelActivate;
            boolean isOverPanelStay     = this.isOverPanelStay;
            boolean isOverPanel         = this.isOverPanel;
            switch(panelState) {
                case PANEL_HIDE:
                    if (isOverPanelActivate) {
                        panelState = PANEL_SHOWING;
                    }
                    break;
                case PANEL_SHOW:
                    if (! isOverPanelStay) {
                        panelState = PANEL_HIDING;
                    }
                    break;
                case PANEL_SHOWING:
                    if (! isOverPanelStay) {
                        panelV = 0;
                        panelState = PANEL_HIDING;
                    } else {
                        int panelDecelDist = PANEL_INV_SLOPE * panelV * panelV /
                                             (2 * PANEL_ACCEL);
                        if (panelX <= panelXLeftmost + panelDecelDist) {
                            panelV -= PANEL_ACCEL;
                            if (panelV <= 0) {
                                panelV = 1;
                            }
                        } else if (panelV < PANEL_MAX_VEL) {
                            panelV += PANEL_ACCEL;
                            if (panelV > PANEL_MAX_VEL) {
                                panelV = PANEL_MAX_VEL;
                            }
                        }
                        panelX -= PANEL_INV_SLOPE * panelV;
                        panelY -= panelV;
                        if (panelX <= panelXLeftmost) {
                            panelX = panelXLeftmost;
                            panelV = 0;
                            panelState = PANEL_SHOW;
                        }
                     }
                    break;
                case PANEL_HIDING:
                    if (isOverPanelStay) {
                        panelV = 0;
                        panelState = PANEL_SHOWING;
                    } else {
                        int panelDecelDist = PANEL_INV_SLOPE * panelV * panelV /
                                             (2 * PANEL_ACCEL);
                        if (panelX >= panelXRightmost - panelDecelDist) {
                            panelV -= PANEL_ACCEL;
                            if (panelV <= 0) {
                                panelV = 1;
                            }
                        } else if (panelV < PANEL_MAX_VEL) {
                            panelV += PANEL_ACCEL;
                            if (panelV > PANEL_MAX_VEL) {
                                panelV = PANEL_MAX_VEL;
                            }
                        }
                        panelX += PANEL_INV_SLOPE * panelV;
                        panelY += panelV;
                        if (panelX >= panelXRightmost) {
                            panelX = panelXRightmost;
                            panelV = 0;
                            for (int i = 0; i < N_CTRLS; i++) {
                                ctrls[i].wasMouseOn = false;
                                ctrls[i].wasMouseOver = false;
                            }
                            panelState = PANEL_HIDE;
                        }
                    }
                    break;
            }
            // Also transparency of panel.
            if (panelState == PANEL_HIDE) {
                incrPanelAlpha(PANEL_ALPHA_INCR);
            } else if (++mouseInactiveCount > PANEL_DIM_COUNT) {
                incrPanelAlpha(-PANEL_ALPHA_DECR);
            } else {
                if (isOverPanel || isOverPanelActivate) {
                    if (panelAlpha < MAX_PANEL_ALPHA) {
                        panelAlpha += PANEL_ALPHA_INCR;
                        if (panelAlpha > MAX_PANEL_ALPHA) {
                            panelAlpha = MAX_PANEL_ALPHA;
                        }
                    }
                } else {
                    if (panelAlpha > MIN_PANEL_ALPHA) {
                        panelAlpha -= PANEL_ALPHA_DECR;
                        if (panelAlpha < MIN_PANEL_ALPHA) {
                            panelAlpha = MIN_PANEL_ALPHA;
                        }
                    }
                }
            }
            
            // Update locations of ctrls.
            for (int i = 0; i < N_CTRLS; i++) {
                ctrls[i].setLocation(panelX + CTRL_X0 + i*CTRL_DX,
                                     panelY + CTRL_Y0 + i*CTRL_DY);
            }
            
            // Composite with panel.
            composite.copy(pixels);
            paintPanel(panelX, panelY, composite);
            // Connection indicator.
            paintConnInd(composite);
            update(composite.getPixels());
            
            t = System.currentTimeMillis();
            try {
                 Thread.sleep(Math.max(DT-t+t0, MIN_SLEEP));
            } catch (InterruptedException e) {}
            
            if (doWait) {
                synchronized(this) {
                    waiting = true;
                }
            }
            
            while (waiting) {
                try {
                     Thread.sleep(WAIT_SLEEP);
                } catch (InterruptedException e) {}
            }
        }
    }
    
    public void stop() {
        super.stop();
        for (int i = 0; i < taps.size(); i++) {
            taps.killTap(i);
        }
        taps.removeAllElements();
        taps = null;
    }
    
    public void kill() {
        if (taps != null) {
            stop();
        }
        evol.kill();
        super.kill();
    }
    
    protected void getApplicationParameters(String[] args) {
        clipsPath = args[0];
        if (clipsPath.charAt(clipsPath.length()-1) != '/') {
            clipsPath = clipsPath + "/";
        }
        for (int i = 1; i < args.length; i++) {
            if (args[i].equalsIgnoreCase("-zoom")) {
                zoom = Integer.parseInt(args[++i]);
            } else if (args[i].equalsIgnoreCase("-wait")) {
                doWait = true;
            } else if (args[i].equalsIgnoreCase("-dbHost")) {
                dbHost = args[++i];
            } else if (args[i].equalsIgnoreCase("-dbUser")) {
                dbUser = args[++i];
            } else if (args[i].equalsIgnoreCase("-dbPass")) {
                dbPass = args[++i];
            } else if (args[i].startsWith("-ctrl")) {
                int j = Integer.parseInt(args[i].substring(5));
                ctrlInit[j] = Integer.parseInt(args[++i]);
            }
        }
    }
    
    protected void getAppletParameters() {
        String s;
        if ((s = getParameter("clipsPath")) != null) {
            clipsPath = s;
        }
        if ((s = getParameter("zoom")) != null) {
            zoom = Integer.parseInt(s);
        }
        if ((s = getParameter("cacheUrl")) != null) {
            cacheUrl = s;
        }
        for (int i = 0; i < N_CTRLS; i++) {
            if ((s = getParameter("ctrl" + i)) != null) {
                ctrlInit[i] = Integer.parseInt(s);
            }
        }
        if (clipsPath.charAt(clipsPath.length()-1) != '/') {
            clipsPath = clipsPath + "/";
        }
    }
    
    protected void initCoords() {
        super.initCoords();
        
        // Adjust for frame.
        appDim.setSize(appDim.width + L + R, appDim.height + T + B);
        origin.translate(L, T);
        
        // Panel coords.
        panelXLeftmost = appDim.width - PANEL_X1 - R;
        panelXRightmost = appDim.width - PANEL_X2 - R;
        panelActivate = 
            new Rectangle(appDim.width - PANEL_ACTIVATE_W - R,
                          PANEL_ACTIVATE_Y + T, 
                          PANEL_ACTIVATE_W, PANEL_ACTIVATE_H);
        panelStay = 
            new Rectangle(appDim.width - PANEL_STAY_W - R, PANEL_STAY_Y + T, 
                          PANEL_STAY_W, PANEL_ACTIVATE_Y + PANEL_ACTIVATE_H);
        panelX = panelXRightmost;
        panelY = panelActivate.y;
        int panelY2 = panelY + panelActivate.height;
        panelXIcept1 = panelX - panelY2 * PANEL_INV_SLOPE;
        panelXIcept2 = panelX - panelY * PANEL_INV_SLOPE;
        panelStayXIcept1 = panelX - (panelY2 + 20) * PANEL_INV_SLOPE;
        panelStayXIcept2 = panelX - (panelY - 20) * PANEL_INV_SLOPE;
        
        erase = new Rectangle(L, T, appDim.width-L-R, appDim.height-T-B);
    }
    
    protected void initCommon() {
        setLayout(null);
        setBackground(Color.black);
        setVisible(true);
        initGfx();
        taps = new Taps();
        if (! isRunningAsApplet) {
            evol = new Evol(dbHost, dbUser, dbPass);
        } else {
            evol = new EvolCache(cacheUrl, this);
        }
        evol.init();
        clipInfo = evol.getClipInfo();
        csFactory = new ClipSegFactory(clipsPath, clipInfo, this);
    }
    
    private void initGfx() {
        MediaTracker mediaTracker = new MediaTracker(this);
        try {
            Image panelImg = getImage("/images/panel.gif");
            mediaTracker.addImage(panelImg, 0);
            Image ticImg = getImage("/images/tic.gif");
            mediaTracker.addImage(ticImg, 0);
            Image ticHiImg = getImage("/images/tic-hi.gif");
            mediaTracker.addImage(ticHiImg, 0);
            Image ticGloImg = getImage("/images/tic-glo.gif");
            mediaTracker.addImage(ticGloImg, 0);
            Image ticSliderImg = getImage("/images/tic-slider.gif");
            mediaTracker.addImage(ticSliderImg, 0);
            
            Image aggCircleImg = getImage("/images/agg-circle.gif");
            mediaTracker.addImage(aggCircleImg, 0);
            Image aggPtsInFlatImg = getImage("/images/agg-pts-in-flat.gif");
            mediaTracker.addImage(aggPtsInFlatImg, 0);
            Image aggPtsInTiltLImg = getImage("/images/agg-pts-in-tilt-l.gif");
            mediaTracker.addImage(aggPtsInTiltLImg, 0);
            Image aggPtsOutFlatImg = getImage("/images/agg-pts-out-flat.gif");
            mediaTracker.addImage(aggPtsOutFlatImg, 0);
            Image aggPtsOutTiltLImg = 
                getImage("/images/agg-pts-out-tilt-l.gif");
            mediaTracker.addImage(aggPtsOutTiltLImg, 0);
            Image aggDiskTopFlatImg = getImage("/images/agg-disk-top-flat.gif");
            mediaTracker.addImage(aggDiskTopFlatImg, 0);
            Image aggDiskTopTiltLImg = 
                getImage("/images/agg-disk-top-tilt-l.gif");
            mediaTracker.addImage(aggDiskTopTiltLImg, 0);
            
            Image[] iconImgs = new Image[N_CTRLS];
            for (int i = 0; i < N_CTRLS; i++) {
                iconImgs[i] = getImage(ICON_IMGS[i]);
                mediaTracker.addImage(iconImgs[i], 0);
            }
            
            Image connIndImg = getImage("/images/conn-ind.gif");
            mediaTracker.addImage(connIndImg, 0);
            
            mediaTracker.waitForID(0);
            
            panelMask = getImageBuffer(panelImg).getMask();
            ticMask = getImageBuffer(ticImg).getMask();
            ticHiMask = getImageBuffer(ticHiImg).getMask();
            ticGloMask = getImageBuffer(ticGloImg).getMask();
            ticSliderMask = getImageBuffer(ticSliderImg).getMask();
            
            aggCircleMask = getImageBuffer(aggCircleImg).getMask();
            aggPtsInFlatMask = getImageBuffer(aggPtsInFlatImg).getMask();
            aggPtsInTiltLMask = getImageBuffer(aggPtsInTiltLImg).getMask();
            aggPtsOutFlatMask = getImageBuffer(aggPtsOutFlatImg).getMask();
            aggPtsOutTiltLMask = getImageBuffer(aggPtsOutTiltLImg).getMask();
            aggDiskTopFlatMask = getImageBuffer(aggDiskTopFlatImg).getMask();
            aggDiskTopTiltLMask = getImageBuffer(aggDiskTopTiltLImg).getMask();
            
            iconMasks = new int[N_CTRLS][];
            for (int i = 0; i < N_CTRLS; i++) {
                iconMasks[i] = getImageBuffer(iconImgs[i]).getMask();
            }
            
            connIndMask = getImageBuffer(connIndImg).getMask();
        } catch (Exception e) {
            System.err.println(e);
        }
        
        // Mirror image masks.
        int h;
        aggPtsInTiltRMask = new int[aggPtsInTiltLMask.length];
        h = aggPtsInTiltLMask.length / AGG_PTS_IN_W;
        PixelDrawer.copyRect(0, 0, AGG_PTS_IN_W, h, 
                             aggPtsInTiltLMask, AGG_PTS_IN_W, h, 
                             0, 0, aggPtsInTiltRMask, AGG_PTS_IN_W, h, 
                             PixelDrawer.COPY_FLIP_HORIZONTALLY);
        aggPtsOutTiltRMask = new int[aggPtsOutTiltLMask.length];
        h = aggPtsOutTiltLMask.length / AGG_PTS_OUT_W;
        PixelDrawer.copyRect(0, 0, AGG_PTS_OUT_W, h, 
                             aggPtsOutTiltLMask, AGG_PTS_OUT_W, h, 
                             0, 0, aggPtsOutTiltRMask, AGG_PTS_OUT_W, h, 
                             PixelDrawer.COPY_FLIP_HORIZONTALLY);
        aggDiskBotFlatMask = new int[aggDiskTopFlatMask.length];
        h = aggDiskTopFlatMask.length / AGG_DISK_W;
        PixelDrawer.copyRect(0, 0, AGG_DISK_W, h, 
                             aggDiskTopFlatMask, AGG_DISK_W, h, 
                             0, 0, aggDiskBotFlatMask, AGG_DISK_W, h, 
                             PixelDrawer.COPY_FLIP_VERTICALLY);
        aggDiskTopTiltRMask = new int[aggDiskTopTiltLMask.length];
        aggDiskBotTiltLMask = new int[aggDiskTopTiltLMask.length];
        aggDiskBotTiltRMask = new int[aggDiskTopTiltLMask.length];
        h = aggDiskTopTiltLMask.length / AGG_DISK_W;
        PixelDrawer.copyRect(0, 0, AGG_DISK_W, h, 
                             aggDiskTopTiltLMask, AGG_DISK_W, h, 
                             0, 0, aggDiskTopTiltRMask, AGG_DISK_W, h, 
                             PixelDrawer.COPY_FLIP_HORIZONTALLY);
        PixelDrawer.copyRect(0, 0, AGG_DISK_W, h, 
                             aggDiskTopTiltLMask, AGG_DISK_W, h, 
                             0, 0, aggDiskBotTiltRMask, AGG_DISK_W, h, 
                             PixelDrawer.COPY_FLIP_VERTICALLY);
        PixelDrawer.copyRect(0, 0, AGG_DISK_W, h, 
                             aggDiskTopTiltRMask, AGG_DISK_W, h, 
                             0, 0, aggDiskBotTiltLMask, AGG_DISK_W, h, 
                             PixelDrawer.COPY_FLIP_VERTICALLY);
        
        ctrls = new Ctrl[N_CTRLS];
        int x = panelX + CTRL_X0;
        int y = panelY + CTRL_Y0;
        int ticH = ticMask.length / TIC_W;
        
        h = CTRL_H + ticH;
        for (int i = 0; i < N_CTRLS; i++) {
            ctrls[i] = new Ctrl(x, y, CTRL_W, h, CTRL_SETTINGS[i]);
            ctrls[i].setting = ctrlInit[i];
            x += CTRL_DX;
            y += CTRL_DY;
        }
        
        // PixelSurfaces.
        if (zoom > 1) {
            pixels = new ZoomPixelSurface(appDim.width, appDim.height, zoom);
            composite = new ZoomPixelSurface(appDim.width, appDim.height, zoom);
        } else {
            pixels = new PixelSurface(appDim.width, appDim.height);
            composite = new PixelSurface(appDim.width, appDim.height);
        }
    }
    
    private void incrPanelAlpha(int n) {
        panelAlpha += n;
        if (panelAlpha > MAX_PANEL_ALPHA) {
            panelAlpha = MAX_PANEL_ALPHA;
        } else if (panelAlpha < MIN_PANEL_ALPHA) {
            panelAlpha = MIN_PANEL_ALPHA;
        }
    }
    
    /* ========================================================================
        PAINT OPS
       ========================================================================
     */
    private void paintBlocks(Clus clus, int fadeFactor, PixelSurface pixels) {
        if (clus.blocks == null) {
            if (Debug.DEBUG) {
                System.err.println("Null blocks in cluster " + clus.clusterID);
            }
            return;
        }
        if (clus.clip == null) {
            // Clip may have been killed tho clus is still around.
            return;
        }
        MPEGFrame frame = clus.clip.getFrame();
        if (frame == null) {
            if (Debug.DEBUG) {
                System.err.println("Frame is null!");
            }
            return;
        }
        if (frameDim == null) {
            frameDim = clus.clip.getFrameDimensions();
            blockDim = clus.clip.getBlockDimensions();
            blockSz = clus.clip.getBlockSize();
        }
        int[] frameRGB = frame.getRGB();
        int tapAlpha = ctrls[TAP0_ALPHA].getValue();
        int mvAlpha = ctrls[MV_ALPHA].getValue();
        if (fadeFactor < 256) {
            tapAlpha = (tapAlpha * fadeFactor) >> 8;
            mvAlpha = (mvAlpha * fadeFactor) >> 8;
        }
        for (int i = 0; i < clus.blocks.length; i++) {
            Block b = clus.blocks[i];
            int x = (b.blockNr % blockDim.width) * blockSz;
            int y = (b.blockNr / blockDim.width) * blockSz;
            if (tapAlpha > 0) {
                pixels.paintRect(x, y, blockSz, blockSz, 
                                 frameRGB, frameDim.width, frameDim.height, 
                                 tapAlpha,
                                 origin.x + (x + b.mv.x) * zoom, 
                                 origin.y + (y + b.mv.y) * zoom);
            }
            // MV
            if (mvAlpha > 0) {
                paintMV(origin.x + (x + 7) * zoom, origin.y + (y + 7) * zoom, 
                        b.mv, mvAlpha, pixels);
            }
        }
    }
    
    private void paintAgg(Clus clus, int rgb, int fadeFactor, 
                          PixelSurface pixels) {
        int x = origin.x + clus.cg.x * zoom;
        int y = origin.y + clus.cg.y * zoom;
        int dx = clus.avgMV.x * zoom;
        int dy = clus.avgMV.y * zoom;
        int alpha = ctrls[AGG_ALPHA].getValue();
        if (fadeFactor < 256) {
            alpha = (alpha * fadeFactor) >> 8;
        }
        int halfAlpha = alpha >> 1;
        
        // Disk top & bottom (indicate avg curl).
        int[] aggDiskTopMask;
        int[] aggDiskBotMask;
        if (clus.avgCurl == 0) {
            aggDiskTopMask = aggDiskTopFlatMask;
            aggDiskBotMask = aggDiskBotFlatMask;
        } else if (clus.avgCurl > 0) {
            aggDiskTopMask = aggDiskTopTiltRMask;
            aggDiskBotMask = aggDiskBotTiltRMask;
        } else {
            aggDiskTopMask = aggDiskTopTiltLMask;
            aggDiskBotMask = aggDiskBotTiltLMask;
        }
        pixels.paintMask(aggDiskTopMask, AGG_DISK_W, 
                         x + AGG_DISK_X, y + AGG_DISK_Y, 
                         AGG_DISK_TOP_RGB, halfAlpha);
        pixels.paintMask(aggDiskBotMask, AGG_DISK_W, 
                         x + AGG_DISK_X, y + AGG_DISK_Y, 
                         AGG_DISK_BOT_RGB, halfAlpha);
        
        // Circle.
        pixels.paintMask(aggCircleMask, AGG_CIRCLE_W, 
                         x + AGG_CIRCLE_X, y + AGG_CIRCLE_Y, rgb, alpha);
        
        // Arrow points (indicate avg div).
        if (clus.avgDiv > 0) {
            int[] aggPtsOutMask = (clus.avgCurl > 0) ? aggPtsOutTiltRMask :
                                  (clus.avgCurl < 0) ? aggPtsOutTiltLMask :
                                                       aggPtsOutFlatMask;
            pixels.paintMask(aggPtsOutMask, AGG_PTS_OUT_W, 
                             x + AGG_PTS_OUT_X, y + AGG_PTS_OUT_Y, 
                             AGG_PTS_RGB, alpha);
        } else if (clus.avgDiv < 0) {
            int[] aggPtsInMask = (clus.avgCurl > 0) ? aggPtsInTiltRMask :
                                 (clus.avgCurl < 0) ? aggPtsInTiltLMask :
                                                      aggPtsInFlatMask;
            pixels.paintMask(aggPtsInMask, AGG_PTS_IN_W, 
                             x + AGG_PTS_IN_X, y + AGG_PTS_IN_Y, 
                             AGG_PTS_RGB, alpha);
        }
        
        // Avg MV.
        pixels.drawWuLine(x, y, x + dx, y + dy, AVG_MV_RGB, alpha);
        pixels.drawWuLine(x+1, y, x+dx+1, y+dy, AVG_MV_RGB, alpha);
        pixels.drawWuLine(x-1, y, x + dx-1, y+dy, AVG_MV_RGB, alpha);
        pixels.drawWuLine(x, y+1, x+dx, y+dy+1, AVG_MV_RGB, alpha);
        pixels.drawWuLine(x, y-1, x+dx, y+dy-1, AVG_MV_RGB, alpha);
        pixels.drawWuLine(x+1, y+1, x+dx+1, y+dy+1, AVG_MV_RGB, halfAlpha);
        pixels.drawWuLine(x+1, y-1, x+dx+1, y+dy-1, AVG_MV_RGB, halfAlpha);
        pixels.drawWuLine(x-1, y+1, x+dx-1, y+dy+1, AVG_MV_RGB, halfAlpha);
        pixels.drawWuLine(x-1, y-1, x+dx-1, y+dy-1, AVG_MV_RGB, halfAlpha);
    }
    
    private void paintPanel(int x, int y, PixelSurface pixels) {
        // Panel backdrop.
        pixels.paintMask(panelMask, PANEL_W, x, y, PANEL_RGB, panelAlpha);
        if (panelState != PANEL_HIDE) {
            // Icons.
            for (int i = 0; i < N_CTRLS; i++) {
            }
            // Base tics.
            int ctrlX = x + CTRL_X0;
            int ctrlY = y + CTRL_Y0;
            int iconX = x + ICON_X0;
            int iconY = y + ICON_Y0;
            int ticX = ctrlX + TIC_X0;
            for (int i = 0; i < N_CTRLS; i++) {
                int ticY = y + CTRL_Y0 + i*CTRL_DY;
                for (int j = 0; j < ctrls[i].settings.length; j++) {
                    pixels.paintMask(ticMask, TIC_W, ticX, ticY, 
                                     TIC_RGB, panelAlpha);
                    ticY += ctrls[i].ticSep;
                }
                // Icon.
                pixels.paintMask(iconMasks[i], ICON_W, iconX, iconY,
                                 TIC_RGB, panelAlpha);
                // Setting.
                int settingPos = ctrls[i].setting * ctrls[i].ticSep;
                pixels.paintMask(ticHiMask, TIC_HI_W, 
                                 ctrlX + TIC_HI_X0,
                                 ctrlY + TIC_HI_Y0 + settingPos,
                                 TIC_HI_RGB, (128 * panelAlpha) >> 8);
                // Glow.
                if (ctrls[i].isMouseOn || ctrls[i].wasMouseOn) {
                    int alpha = (panelAlpha < 255) ? 
                                    (panelAlpha * ctrls[i].alpha) >> 8 :
                                    ctrls[i].alpha;
                    pixels.paintMask(ticGloMask, TIC_GLO_W, 
                                     ctrlX + TIC_GLO_X0,
                                     ctrlY + TIC_GLO_Y0 + settingPos,
                                     TIC_GLO_RGB, alpha);
                    pixels.paintMask(ticSliderMask, TIC_GLO_W, 
                                     ctrlX + TIC_GLO_X0,
                                     ctrlY + TIC_GLO_Y0 + settingPos,
                                     TIC_GLO_RGB, alpha);
                    pixels.paintMask(ticHiMask, TIC_HI_W, 
                                     ctrlX + TIC_HI_X0,
                                     ctrlY + TIC_HI_Y0 + settingPos,
                                     TIC_HI_RGB, alpha);
                    if (! ctrls[i].isMouseOn) {
                        ctrls[i].fadeAlpha();
                    }
                } else if (ctrls[i].isMouseOver || ctrls[i].wasMouseOver) {
                    int alpha = (panelAlpha < 255) ? 
                                    (panelAlpha * ctrls[i].alpha) >> 8 :
                                    ctrls[i].alpha;
                    pixels.paintMask(ticGloMask, TIC_GLO_W, 
                                     ctrlX + TIC_GLO_X0,
                                     ctrlY + TIC_GLO_Y0 + settingPos,
                                     TIC_GLO_RGB, alpha);
                    pixels.paintMask(ticHiMask, TIC_HI_W, 
                                     ctrlX + TIC_HI_X0,
                                     ctrlY + TIC_HI_Y0 + settingPos,
                                     TIC_HI_RGB, alpha);
                    if (! ctrls[i].isMouseOver) {
                        ctrls[i].fadeAlpha();
                    }
                }
                if (ctrls[i].isMouseOn || ctrls[i].wasMouseOn ||
                    ctrls[i].isMouseOver || ctrls[i].wasMouseOver) {
                    int alpha = (panelAlpha < 255) ? 
                                    (panelAlpha * ctrls[i].alpha) >> 8 :
                                    ctrls[i].alpha;
                    pixels.paintMask(iconMasks[i], ICON_W, iconX, iconY,
                                     TIC_GLO_RGB, alpha);
                }
                ctrlX += CTRL_DX;
                ctrlY += CTRL_DY;
                iconX += CTRL_DX;
                iconY += CTRL_DY;
                ticX += CTRL_DX;
            }
        }
    }
    
    private void paintConnInd(PixelSurface pixels) {
        if (connIndMask == null) {
            return;
        }
        pixels.paintMask(connIndMask, CONN_IND_W, 
                         appDim.width-CONN_IND_W-1, appDim.height-CONN_IND_H-1, 
                         CONN_RGB[connStatus], 208);
    }
    
    private void paintErrorMsg(PixelSurface pixels) {
        if (errMsgMask == null) {
            MediaTracker mediaTracker = new MediaTracker(this);
            try {
                Image errMsgImg = getImage(getCodeBase(), "error-msg.gif");
                mediaTracker.addImage(errMsgImg, 0);
                mediaTracker.waitForID(0);
                errMsgW = errMsgImg.getWidth(this);
                errMsgMask = getImageBuffer(errMsgImg).getMask();
            } catch (Exception e) {}
        }
        if (errMsgMask != null) {
            pixels.paintMask(errMsgMask, errMsgW, origin.x, origin.y, 
                             ERR_MSG_RGB, 255);
        }
    }
    
    /* ========================================================================
        MOUSE OPS
       ========================================================================
     */
    public void mouseMoved(MouseEvent e) {
        Point pt = e.getPoint();
        e.consume();
        checkMousePos(pt);
        mouseInactiveCount = 0;
    }
    
    public void mousePressed(MouseEvent e) {
        Point pt = e.getPoint();
        e.consume();
        checkMousePos(pt);
        mouseInactiveCount = 0;
        for (int i = 0; i < N_CTRLS; i++) {
            ctrls[i].isMouseOn = ctrls[i].isMouseOver;
            if (ctrls[i].isMouseOn) {
                ctrls[i].alpha = 255;
                // Move to closest tic.
                ctrls[i].moveSetting(pt);
            }
            if (! ctrls[i].wasMouseOn) {
                ctrls[i].wasMouseOn = ctrls[i].isMouseOn;
            }
        }
    }
    
    public void mouseDragged(MouseEvent e) {
        Point pt = e.getPoint();
        e.consume();
        mouseInactiveCount = 0;
        for (int i = 0; i < N_CTRLS; i++) {
            if (! ctrls[i].isMouseOn) {
                continue;
            }
            ctrls[i].alpha = 255;
            // Move to closest tic.
            ctrls[i].moveSetting(pt);
        }
    }
    
    public void mouseReleased(MouseEvent e) {
        mouseInactiveCount = 0;
        for (int i = 0; i < N_CTRLS; i++) {
            if (ctrls[i].isMouseOn) {
                ctrls[i].wasMouseOn = true;
                ctrls[i].isMouseOn = false;
            }
        }
    }
    
    public void mouseEntered(MouseEvent e) {}
    
    public void mouseExited(MouseEvent e) {}
    
    public void mouseClicked(MouseEvent e) {}
    
    private void checkMousePos(Point pt) {
        isOverPanelActivate = panelActivate.contains(pt);
        isOverPanelStay = panelStay.contains(pt) &&
                          pt.x > PANEL_INV_SLOPE * pt.y + panelStayXIcept1 &&
                          pt.x < PANEL_INV_SLOPE * pt.y + panelStayXIcept2;
        isOverPanel = pt.x >= panelX &&
                      pt.y >= panelY &&
                      pt.x > PANEL_INV_SLOPE * pt.y + panelXIcept1 &&
                      pt.x < PANEL_INV_SLOPE * pt.y + panelXIcept2;
        for (int i = 0; i < N_CTRLS; i++) {
            ctrls[i].isMouseOver = ctrls[i].contains(pt);
            if (ctrls[i].isMouseOver) {
                ctrls[i].alpha = 255;
            }
            if (! ctrls[i].wasMouseOver) {
                ctrls[i].wasMouseOver = ctrls[i].isMouseOver;
            }
        }
    }
    
    /* ========================================================================
        PRIVATE CLASSES
       ========================================================================
     */
    final private class Ctrl extends Rectangle {
        
        final static int ALPHA_DECR = 10;
        
        int[] settings;
        int ticSep;
        int setting;
        int alpha;
        boolean isMouseOver;
        boolean isMouseOn;
        boolean wasMouseOver;
        boolean wasMouseOn;
        
        Ctrl(int x, int y, int width, int height, int[] settings) {
            super(x, y, width, height);
            this.settings = settings;
            this.ticSep = CTRL_H / (settings.length-1);
        }
        
        void fadeAlpha() {
            alpha -= ALPHA_DECR;
            if (alpha <= 0) {
                alpha = 0;
                wasMouseOver = false;
                wasMouseOn = false;
            }
        }
        
        void moveSetting(Point pt) {
            int xOff = pt.x - x - TIC_X0;
            int yOff = pt.y - y + 
                       Math.round((float) (ticSep-1) / 2.0f -
                                  (float) xOff / (float) PANEL_INV_SLOPE);
            setting = yOff / ticSep;
            if (setting >= settings.length) {
                setting = settings.length-1;
            } else if (setting < 0) {
                setting = 0;
            }
        }
        
        int getValue() {
            return settings[setting];
        }
    }
}
